from pydesy.gmath import *


class Point:
    def __init__(self, x: float = 0, y: float = 0, z: float = 0):
        self.x = x
        self.y = y
        self.z = z

    def get_dict(self):
        return {"x": self.x, "y": self.y, "h": self.z}

    def get_list_h(self):
        return [self.x, self.y, self.z]

    def get_list_v(self):
        return [
            [self.x],
            [self.y],
            [self.z]
        ]

    def from_list(self, array):
        self.x, self.y, self.z = array[0], array[1], array[2]

    def dgt(self, d: float, a: float):
        return direct_geo_task(self, d, a)

    def igt(self, point) -> list[float]:
        return inverse_geo_task(self, point)

    def dist2(self, point) -> float:
        return distance2d(self, point)

    def dist3(self, point) -> float:
        return distance3d(self, point)

    def reverse(self):
        self.x, self.y = self.y, self.x

    def __add__(self, other):
        if isinstance(other, Point):
            return Point(other.x + self.x, other.y + self.y, other.z + self.z)
        else:
            return NotImplemented

    def __mul__(self, other):
        if isinstance(other, Point):
            return Point(other.x * self.x, other.y * self.y, other.z * self.z)

        elif isinstance(other, int) or isinstance(other, float):
            return Point(other * self.x, other * self.y, other * self.z)

        else:
            return NotImplemented

    def __pow__(self, power, modulo=None):
        return Point(self.x ** power, self.y ** power, self.z ** power)

    def __neg__(self):
        return Point(-self.x, -self.y, -self.z)

    def __pos__(self):
        return Point(self.x, self.y, self.z)

    def __repr__(self):
        return f"<x: {self.x}, y: {self.y}, h: {self.z}>"

    def __str__(self):
        return f"<x: {self.x}, y: {self.y}, h: {self.z}>"


# Direct geodetic task
def direct_geo_task(point: Point, d, a) -> Point:
    return Point(point.x + d * cos(a), point.y + d * sin(a))


# Inverse geodetic task
def inverse_geo_task(point1: Point, point2: Point) -> list[float]:
    dx, dy = point2.x - point1.x, point2.y - point1.y
    d = point1.dist2(point2)
    r = atan(abs(dy / dx))

    a = 0
    if dx > 0 and dy > 0:
        a = r
    elif dx < 0 < dy:
        a = 180 - r
    elif dx < 0 and dy < 0:
        a = 180 + r
    elif dx > 0 > dy:
        a = 360 - r
    elif dx == 0:
        if dy > 0:
            a = 90
        elif dy < 0:
            a = 270
    elif dy == 0:
        if dx > 0:
            a = 0
        elif dx < 0:
            a = 180
    else:
        a = None

    return d, a


# Horizontal distance
def horizontal_distance(s: float = 0, v: float = 0) -> float:
    return s * cos(v)

# Distance between Point1 and Point2
def distance2d(point1: Point, point2: Point) -> float:
    return math.sqrt((point2.x - point1.x) ** 2 + (point2.y - point1.y) ** 2)

def distance3d(point1: Point, point2: Point) -> float:
    return math.sqrt((point2.x - point1.x) ** 2 + (point2.y - point1.y) ** 2 + (point2.z - point1.z) ** 2)


class TLE:
    def __init__(self, line1 : str, line2 : str):
        spline1 = line1.split()
        spline2 = line2.split()

        #Line 1
        self.satellite_number = int(spline1[1][:-1])
        self.classification = spline1[1][-1:]
        self.launch_year = int(spline1[2][:2])
        self.launch_number = int(spline1[2][2:5])
        self.launch_piece = spline1[2][-1:]
        self.epoch_year = int(spline1[3][:2])
        self.epoch = float(spline1[3][2:])
        self.first_derivative = float(("-0" if spline1[4][0] == "-" else "0") + spline1[4][1:])
        self.second_derivative = float(("-0." if spline1[5][0] == "-" else "0.") + spline1[5][1:].replace("-", "e-"))
        self.BSTAR = float(("-0." if spline1[6][0] == "-" else "0.") + spline1[6][1:].replace("-", "e-"))
        self.ephemeris_type = spline1[7]
        self.element_number = int(spline1[8][:-1])
        self.checksum1 = int(spline1[8][-1:])

        #Line 2
        self.inclination = float(spline2[2])
        self.right_ascension = float(spline2[3])
        self.eccentricity = float("0." + spline2[4])
        self.perigee = float(spline2[5])
        self.mean_anomaly = float(spline2[6])
        self.mean_motion = float(spline2[7][:-6])
        self.revolution_number = int(spline2[7][-6:-1])
        self.checksum2 = int(spline2[7][-1])

    def print(self):
        for var in list(vars(self)):
            if type(var) == int or float or str:
                print(f"{var.replace('_', ' ').title()} = {vars(self)[var]}")


import matplotlib.pyplot as plt

from pydesy.basic import *
from pydesy.gmath import *
from pydesy.geometry import *

class PolarSerif:
    def __init__(self, A, B):
        self.A = A
        self.B = B
        self.directionAngle = A.igt(B)[1]

    def set_side_lengths(self, sideLengths: list[float]):
        self.sideLengths = sideLengths

    def set_horizontal_angles(self, horizontalAngles: list[float]):
        self.horizontalAngles = horizontalAngles

    def set_points(self, points: list):
        self.points = points

    def calculate_serif(self) -> list[Point]:
        self.points = list()
        for i in range(len(self.sideLenghts)):
            directionAngle = revs(self.directionAngle + self.horizontalAngles[i])
            print(directionAngle)
            dx = self.sideLengths[i] * cos(directionAngle)
            dy = self.sideLengths[i] * sin(directionAngle)
            self.points.append(Point(self.A.x + dx, self.A.y + dy))
        return self.points

    def calculate_accuracy_score(self, lenghtError: float, angleError: float) -> list[float]:
        self.errors = list()
        for i in range(len(self.points)):
            self.errors.append(self.sideLengths[i] ** 2 * angleError ** 2 / r ** 2 + lenghtError ** 2)
        return self.errors

    def draw_serif(self):
        plt.figure("Polar serif")
        plt.text(self.A.y, self.A.x, "A")
        plt.text(self.B.y, self.B.x, "B")
        for i, point in enumerate(self.points):
            plt.plot([self.A.y, point.y], [self.A.x, point.x], linestyle="--", color="blue", marker=".")
            plt.text(point.y, point.x, f"{i}")

        plt.plot([self.A.y, self.B.y], [self.A.x, self.B.x], color="red", marker="^")
        plt.xlabel("y")
        plt.ylabel("x")
        plt.title("Polar serif")
        plt.show()


class LinearSerif:
    def __init__(self, points: list[Point]):
        self.points = points

    def set_side_lengths(self, sideLengths: list[float]):
        self.sideLengths = sideLengths

    def calculate_serif(self) -> Point:
        self.calculatedPoints = list()
        self.gammas = list()
        for i in range(1, len(self.points)):
            basis, directionAngle = self.points[i-1].igt(self.points[i])
            gamma, sigma, betta = three_sides(basis, self.sideLengths[i-1], self.sideLengths[i])
            correctedDirectionAngle = revs(directionAngle - betta)
            self.calculatedPoints.append(self.points[i-1].dgt(self.sideLengths[i-1], correctedDirectionAngle))
            self.gammas.append(gamma)
            if i == len(self.points):
                correctedDirectionAngle = revs(directionAngle + sigma)
                self.calculatedPoints.append(self.points[i].dgt(self.sideLengths[i], correctedDirectionAngle))

        self.calculatedPoint = Point()
        for point in self.calculatedPoints:
            self.calculatedPoint += point * (1 / len(self.calculatedPoints))
        return self.calculatedPoint

    def calculate_accuracy_score(self, lenghtError: float) -> list[float]:
        self.errors = list()
        for gamma in self.gammas:
            self.errors.append(lenghtError * sqrt(2) / sin(gamma))
        return sum(self.errors) / len(self.errors)

    def draw_serif(self):
        plt.figure("Linear serif")
        plt.text(self.calculatedPoint.y, self.calculatedPoint.x, "?")
        for i, point in enumerate(self.points):
            plt.plot([point.y, self.calculatedPoint.y], [point.x, self.calculatedPoint.x], linestyle="--", color="blue", marker=".")
            plt.text(point.y, point.x, f"{i}")
        for i in range(1, len(self.points)):
            plt.plot([self.points[i-1].y, self.points[i].y], [self.points[i-1].x, self.points[i].x], color="red", marker="^")
        plt.xlabel("y")
        plt.ylabel("x")
        plt.title("Linear serif")
        plt.show()


from pydesy.gmath import *
from pydesy.basic import *
def polygon_area(polygon : list) -> float:
    countPolygons = len(polygon)    
    if countPolygons < 3:
        raise Exception("To solve this problem, you need at least 3 points")
        
    firstSum = sum([polygon[i].x * polygon[i+1].y for i in range(countPolygons - 1)]) +  polygon[countPolygons - 1].x * polygon[0].y
    secondSum = sum([polygon[i+1].x * polygon[i].y for i in range(countPolygons - 1)]) +  polygon[countPolygons - 1].y * polygon[0].x
    return (firstSum - secondSum) / 2

def three_sides(a : float, b : float, c : float) -> list[float]:
    A = acos((b ** 2 + c ** 2 - a ** 2) / (2 * b * c))
    B = acos((a ** 2 + c ** 2 - b ** 2) / (2 * a * c))
    C = acos((a ** 2 + b ** 2 - c ** 2) / (2 * a * b))

    return [A, B, C]


#Сторонние библиотеки
import matplotlib.pyplot as plt

from pydesy.gmath import *
from pydesy.basic import *

from enum import Enum
from datetime import datetime

class Scales(Enum):
    S1_5000 = 1
    S1_2000 = 2
    S1_1000 = 3
    S1_500 = 4

class TheodoliteRequirement:
    requirement = {
        Scales.S1_5000 : {
            "maximum_length" : 1200,
            "maximum_number_of_sides" : 6,
            "limiting_side_length" : 300,
            "limit_relative_error": 1 / 2000,
            "limit_horizontal_discrepancy": lambda x: 1 * sqrt(x)
        },
        
        Scales.S1_2000 : {
            "maximum_length" : 600,
            "maximum_number_of_sides" : 5,
            "limiting_side_length" : 200,
            "limit_relative_error": 1 / 2000,
            "limit_horizontal_discrepancy": lambda x: 1 * sqrt(x)
        },
        
        Scales.S1_1000 : {
            "maximum_length" : 300,
            "maximum_number_of_sides" : 3,
            "limiting_side_length" : 150,
            "limit_relative_error": 1 / 2000,
            "limit_horizontal_discrepancy": lambda x: 1 * sqrt(x)
        },
        
        Scales.S1_500 : {
            "maximum_length" : 100,
            "maximum_number_of_sides" : 2,
            "limiting_side_length" : 100,
            "limit_relative_error": 1 / 2000,
            "limit_horizontal_discrepancy": lambda x: 1 * sqrt(x)
        }
        
    }
    
    def setScaleRequirement(typeScale, values):
        TheodoliteRequirement.requirement[typeScale] = values
                                          
class TheodoliteCourse: 
    def __init__(self, scale=Scales.S1_500):
        self.scale = scale
    
    def set_scale(self, scale : float):
        self.scale = scale
    
    def set_points(self, startPoint : Point, endPoint : Point):
        self.startPoint =  startPoint
        self.endPoint = endPoint
        
    def set_direction_angles(self, startDirectionAngle : list, endDirectionAngle : list):
        self.startDirectionAngle = startDirectionAngle
        self.endDirectionAngle = endDirectionAngle

    def set_horizontal_angles(self, horizontalAngles : list):
        self.horizontalAngles = horizontalAngles
        
    def set_side_lenghts(self, sideLenghts : list):
        self.sideLenghts = sideLenghts
        
    def set_average_excess(self, averageExcess : list):
        self.averageExcess = averageExcess
        
    def calculate_course(self) -> dict:
        self.errors = []
        self.horizontalDiscrepancy = (angle(sum(self.horizontalAngles)) - angle(self.endDirectionAngle - self.startDirectionAngle) - 180) * 60
        self.limitHorizontalDiscrepancy = TheodoliteRequirement.requirement[self.scale]["limit_horizontal_discrepancy"](len(self.horizontalAngles))
        
        if abs(self.horizontalDiscrepancy) > self.limitHorizontalDiscrepancy:
            self.errors.append(f"Horizontal angle residual is exceeded residual limit: {abs(self.horizontalDiscrepancy)} > {self.limitHorizontalDiscrepancy}")

        self.sumSideLenghts = sum(self.sideLenghts)
        if self.sumSideLenghts > TheodoliteRequirement.requirement[self.scale]["maximum_length"]:
            self.errors.append(f"The sum lengths of the theodolite course is greater than the specified value: {self.sumSideLenghts} > {TheodoliteRequirement.requirement[self.scale]['maximum_length']}")

        self.angleCorrection = -self.horizontalDiscrepancy / len(self.horizontalAngles)
        self.directionAngles = [angle(self.startDirectionAngle + 180 + angle(self.horizontalAngles[0], s = self.angleCorrection))]
        self.incrementsX, self.incrementsY = list(), list()
        for i in range(1, len(self.horizontalAngles)):
            self.directionAngles.append(angle(self.directionAngles[i-1] + 180 + angle(self.horizontalAngles[i], s = self.angleCorrection)))
            self.incrementsX.append(-cos(self.directionAngles[i-1]) * self.sideLenghts[i-1])
            self.incrementsY.append(-sin(self.directionAngles[i-1]) * self.sideLenghts[i-1])

        self.incrementXDiscrepancy = sum(self.incrementsX) - (self.endPoint.x - self.startPoint.x)
        self.incrementYDiscrepancy = sum(self.incrementsY) - (self.endPoint.y - self.startPoint.y)
        self.incrementXY = sqrt(self.incrementXDiscrepancy ** 2 + self.incrementYDiscrepancy ** 2)
        self.relativeError = self.incrementXY / self.sumSideLenghts
        
        if  self.relativeError > TheodoliteRequirement.requirement[self.scale]["limit_relative_error"]:
            self.errors.append(f"The relative distance error is greater than the allowable value: {self.relativeError} > {TheodoliteRequirement.requirement[self.scale]['limit_relative_error']}")
        
        self.incrementXCorrections = [-self.incrementXDiscrepancy * self.sideLenghts[i] / self.sumSideLenghts for i in range(len(self.sideLenghts))]
        self.incrementYCorrections = [-self.incrementYDiscrepancy * self.sideLenghts[i] / self.sumSideLenghts for i in range(len(self.sideLenghts))]
        self.points = [Point(self.startPoint.x + self.incrementsX[0] + self.incrementXCorrections[0], self.startPoint.y + self.incrementsY[0] + self.incrementYCorrections[0])]
        for i in range(1, len(self.sideLenghts)-1):
            self.points.append(Point(self.points[i-1].x + self.incrementsX[i] + self.incrementXCorrections[i], self.points[i-1].y + self.incrementsY[i] + self.incrementYCorrections[i]))
            if self.sideLenghts[i-1] > TheodoliteRequirement.requirement[self.scale]["limiting_side_length"]:
                self.errors.append(f"Side length is greater than the specified value: {self.sideLenghts[i-1]} > {TheodoliteRequirement.requirement[self.scale]['limiting_side_length']}")
        self.averageExcessDiscrepancy = sum(self.averageExcess) - (self.endPoint.h - self.startPoint.h)
        self.limitAverageExcessDiscrepancy = (0.04 * self.sumSideLenghts) / (100 * sqrt(len(self.sideLenghts)))

        if len(self.sideLenghts) > TheodoliteRequirement.requirement[self.scale]["maximum_number_of_sides"]:
            self.errors.append(f"Sides count is greater than the specified value: {len(self.sideLenghts)} > {TheodoliteRequirement.requirement[self.scale]['maximum_number_of_sides']}")

        if abs(self.averageExcessDiscrepancy) > self.limitAverageExcessDiscrepancy:
            self.errors.append(f"The discrepancy of excess is greater than the admissible discrepancy: {abs(self.averageExcessDiscrepancy)} > {self.limitAverageExcessDiscrepancy}")
        
        self.averageExcessCorrections = [-self.averageExcessDiscrepancy * self.sideLenghts[i] / self.sumSideLenghts for i in range(len(self.sideLenghts))]
        self.points[0].z = self.startPoint.h + self.averageExcess[0] + self.averageExcessCorrections[0]
        for i in range(1, len(self.averageExcess)-1):
            self.points[i].z = self.points[i-1].z + self.averageExcess[i] + self.averageExcessCorrections[i]
        return {
            "points" : self.points,
            "relativeError" : self.relativeError,
            "errors" : self.errors
        }
    
    def save_report(self, fileName : str):
        with open(fileName, "w") as report:
            report.write(f"Report at {datetime.now()}\n")
            report.write("Points:\n\tName\tX\tY\tH\n")
            report.write(f"\tstart\t{self.startPoint.x}\t{self.startPoint.y}\t{self.startPoint.h}\n")
            for i, point in enumerate(self.points):
                report.write(f"\t{i}\t{point.x}\t{point.y}\t{point.z}\n")
            report.write(f"\tend\t{self.endPoint.x}\t{self.endPoint.y}\t{self.endPoint.h}\n")
            report.write(f"Relative distance error: {self.relativeError}\n")
            report.write(f"Errors:\n\tID\tText\n")
            for i, error in enumerate(self.errors):
                report.write(f"\t{i}\t{error}\n")

    def draw_cource(self):
        plt.figure("Theodolite course")
        plt.text(self.startPoint.y, self.startPoint.x,  f"start, h: {self.startPoint.h}")
        plt.text(self.endPoint.y, self.endPoint.x,  f"end, h: {self.endPoint.h}")
        for i, point in enumerate(self.points):
            plt.text(point.y, point.x, f"{i}, h: {point.z}")

        plt.plot([self.startPoint.y] + [point.y for point in self.points] +  [self.endPoint.y], [self.startPoint.x] + [point.x for point in self.points] + [self.endPoint.x], marker=".")
        plt.xlabel("y")
        plt.ylabel("x")
        plt.title("Theodolite course")
        plt.show()


from setuptools import setup

setup(name='pydesy',
      version='0.0.5',
      description='Geodetic library',
      packages=['pydesy'],
      author='Maxim Shmakov',
      author_email='shmakov.maxim2001@yandex.ru',
      install_requires=['numpy', 'matplotlib'],
      url='https://github.com/ChevisOfficial/pydesy',
      zip_safe=False)
